@@ -16,9 +16,10 @@ Expressions, including loops, conditionals, and function definitions can be
 compiled into ast node and then evaluated later, using the current values
 in the symbol table.
 
-The result is a restricted, simplified version of Python meant for
-numerical calculations that is somewhat safer than 'eval' because many
-unsafe operations (such as 'import' and 'eval') are simply not allowed.
+The result is a restricted, simplified version of Python meant for numerical
+calculations that is somewhat safer than 'eval' because many unsafe operations
+(such as 'eval') are simply not allowed, and others (such as 'import') are
+disabled by default, but can be explicitly enabled.
 
 Many parts of Python syntax are supported, including:
      for loops, while loops, if-then-elif-else conditionals, with,
@@ -34,7 +35,7 @@ The following Python syntax elements are not supported:
 
 In addition, while many builtin functions are supported, several builtin
 functions that are considered unsafe are missing ('eval', 'exec', and
-'getattr' for example)
+'getattr' for example) are missing.
 """
 import ast
 import sys
@@ -43,7 +44,7 @@ import inspect
 import time
 from sys import exc_info, stderr, stdout
 
-from .astutils import (HAS_NUMPY, UNSAFE_ATTRS, ExceptionHolder, ReturnedNone,
+from .astutils import (HAS_NUMPY, UNSAFE_ATTRS, ExceptionHolder, ReturnedNone, Empty,
                        make_symbol_table, numpy, op2func, valid_symbol_name,
                        Procedure)
 
@@ -61,11 +62,11 @@ ALL_NODES = ['arg', 'assert', 'assign', 'attribute', 'augassign', 'binop',
 MINIMAL_CONFIG = {'import': False, 'importfrom': False}
 DEFAULT_CONFIG = {'import': False, 'importfrom': False}
 
-for node in ('assert', 'augassign', 'delete', 'if', 'ifexp', 'for',
+for _tnode in ('assert', 'augassign', 'delete', 'if', 'ifexp', 'for',
              'formattedvalue', 'functiondef', 'print', 'raise', 'listcomp',
              'dictcomp', 'setcomp', 'try', 'while', 'with'):
-    MINIMAL_CONFIG[node] = False
-    DEFAULT_CONFIG[node] = True
+    MINIMAL_CONFIG[_tnode] = False
+    DEFAULT_CONFIG[_tnode] = True
 
 class Interpreter:
     """create an asteval Interpreter: a restricted, simplified interpreter
@@ -74,8 +75,10 @@ class Interpreter:
     Parameters
     ----------
     symtable : dict or `None`
-        dictionary to use as symbol table (if `None`, one will be created).
-    usersyms : dict or `None`
+        dictionary or SymbolTable to use as symbol table (if `None`, one will be created).
+    nested_symtable : bool, optional
+        whether to use a new-style nested symbol table instead of a plain dict [False]
+    user_symbols : dict or `None`
         dictionary of user-defined symbols to add to symbol table.
     writer : file-like or `None`
         callable file-like object where standard output will be sent.
@@ -102,14 +105,21 @@ class Interpreter:
        'assert', 'delete', 'raise', 'print')
     2. by default 'import' and 'importfrom' are disabled, though they can be enabled.
     """
-    def __init__(self, symtable=None, usersyms=None, writer=None,
-                 err_writer=None, use_numpy=True, max_statement_length=50000,
-                 minimal=False, readonly_symbols=None, builtins_readonly=False,
-                 config=None, **kws):
+    def __init__(self, symtable=None, nested_symtable=False,
+                 user_symbols=None, writer=None, err_writer=None,
+                 use_numpy=True, max_statement_length=50000,
+                 minimal=False, readonly_symbols=None,
+                 builtins_readonly=False, config=None, **kws):
 
         self.config = copy.copy(MINIMAL_CONFIG if minimal else DEFAULT_CONFIG)
         if config is not None:
             self.config.update(config)
+        self.config['nested_symtable'] = nested_symtable
+
+        if user_symbols is None:
+            user_symbols = {}
+            if 'usersyms' in kws:
+                user_symbols = kws.pop('usersyms') # back compat, changed July, 2023, v 0.9.4
 
         if len(kws) > 0:
             for key, val in kws.items():
@@ -126,10 +136,10 @@ class Interpreter:
         self.err_writer = err_writer or stderr
         self.max_statement_length = max(1, min(1.e8, max_statement_length))
 
+        self.use_numpy = HAS_NUMPY and use_numpy
         if symtable is None:
-            if usersyms is None:
-                usersyms = {}
-            symtable = make_symbol_table(use_numpy=use_numpy, **usersyms)
+            symtable = make_symbol_table(nested=nested_symtable,
+                                         use_numpy=self.use_numpy, **user_symbols)
 
         symtable['print'] = self._printer
         self.symtable = symtable
@@ -141,7 +151,6 @@ class Interpreter:
         self._calldepth = 0
         self.lineno = 0
         self.start_time = time.time()
-        self.use_numpy = HAS_NUMPY and use_numpy
 
         self.node_handlers = {}
         for node in ALL_NODES:
@@ -333,7 +342,7 @@ class Interpreter:
                 raise exc(errmsg)
             if show_errors:
                 print(errmsg, file=self.err_writer)
-
+        return None
 
     @staticmethod
     def dump(node, **kw):
@@ -430,10 +439,6 @@ class Interpreter:
         """Pass statement."""
         return None  # ()
 
-    def on_ellipsis(self, node):
-        """Ellipses.  deprecated in 3.8"""
-        return Ellipsis
-
     # for break and continue: set the instance variable _interrupt
     def on_interrupt(self, node):    # ()
         """Interrupt handler."""
@@ -451,9 +456,8 @@ class Interpreter:
     def on_assert(self, node):    # ('test', 'msg')
         """Assert statement."""
         if not self.run(node.test):
-            #depraction warning: will become:
-            #  msg = node.msg.value if node.msg else ""
-            msg = node.msg.s if node.msg else ""
+            msg = node.msg.value if node.msg else ""
+            # msg = node.msg.s if node.msg else ""
             self.raise_exception(node, exc=AssertionError, msg=msg)
         return True
 
@@ -478,18 +482,6 @@ class Interpreter:
         """Return constant value."""
         return node.value
 
-    def on_num(self, node):   # ('n',)
-        """Return number.  deprecated in 3.8"""
-        return node.n
-
-    def on_str(self, node):   # ('s',)
-        """Return string.  deprecated in 3.8"""
-        return node.s
-
-    def on_bytes(self, node):
-        """return bytes.  deprecated in 3.8"""
-        return node.s  # ('s',)
-
     def on_joinedstr(self, node):  # ('values',)
         "join strings, used in f-strings"
         return ''.join([self.run(k) for k in node.values])
@@ -505,19 +497,19 @@ class Interpreter:
             fmt = f'{{0:{self.run(node.format_spec)}}}'
         return fmt.format(val)
 
+    def _getsym(self, node):
+        val = self.symtable.get(node.id, ReturnedNone)
+        if isinstance(val, Empty):
+            msg = f"name '{node.id}' is not defined"
+            self.raise_exception(node, exc=NameError, msg=msg)
+        return val
+
     def on_name(self, node):    # ('id', 'ctx')
         """Name node."""
         ctx = node.ctx.__class__
         if ctx in (ast.Param, ast.Del):
             return str(node.id)
-        if node.id in self.symtable:
-            return self.symtable[node.id]
-        msg = f"name '{node.id}' is not defined"
-        self.raise_exception(node, exc=NameError, msg=msg)
-
-    def on_nameconstant(self, node):
-        """True, False, or None  deprecated in 3.8"""
-        return node.value
+        return self._getsym(node)
 
     def node_assign(self, node, val):
         """Assign a value (not the node.value object) to a node.
@@ -575,6 +567,7 @@ class Interpreter:
         # AttributeError or accessed unsafe attribute
         msg = f"no attribute '{node.attr}' for {self.run(node.value)}"
         self.raise_exception(node, exc=AttributeError, msg=msg)
+        return None
 
     def on_assign(self, node):    # ('targets', 'value')
         """Simple assignment."""
@@ -608,6 +601,7 @@ class Interpreter:
             return val[nslice]
         msg = "subscript with unknown context"
         self.raise_exception(node, msg=msg)
+        return None
 
     def on_delete(self, node):    # ('targets',)
         """Delete statement."""
@@ -655,15 +649,18 @@ class Interpreter:
             rval = self.run(rnode)
             ret = op2func(oper)(lval, rval)
             results.append(ret)
-            if ((self.use_numpy and not isinstance(ret, numpy.ndarray)) and
-                    not ret):
-                break
+            try:
+                if not ret:
+                    break
+            except ValueError:
+                pass
             lval = rval
         if len(results) == 1:
-            return results[0]
-        out = True
-        for ret in results:
-            out = out and ret
+            out = results[0]
+        else:
+            out = True
+            for ret in results:
+                out = out and ret
         return out
 
     def _printer(self, *out, **kws):
@@ -745,78 +742,64 @@ class Interpreter:
             if hasattr(ctx, '__exit__'):
                 ctx.__exit__()
 
-
-    def comprehension_data(self, node):      # ('elt', 'generators')
-        "data for comprehensions"
-        mylocals = {}
+    def _comp_save_syms(self, node):
+        """find and save symbols that will be used in a comprehension"""
         saved_syms = {}
-
         for tnode in node.generators:
-            if tnode.__class__ == ast.comprehension:
-                if tnode.target.__class__ == ast.Name:
-                    if (not valid_symbol_name(tnode.target.id) or
-                        tnode.target.id in self.readonly_symbols):
-                        errmsg = f"invalid symbol name (reserved word?) {tnode.target.id}"
-                        self.raise_exception(tnode.target, exc=NameError, msg=errmsg)
-                    mylocals[tnode.target.id] = []
-                    if tnode.target.id in self.symtable:
-                        saved_syms[tnode.target.id] = copy.deepcopy(self.symtable[tnode.target.id])
+            if tnode.target.__class__ == ast.Name:
+                if (not valid_symbol_name(tnode.target.id) or
+                    tnode.target.id in self.readonly_symbols):
+                    errmsg = f"invalid symbol name (reserved word?) {tnode.target.id}"
+                    self.raise_exception(tnode.target, exc=NameError, msg=errmsg)
+                if tnode.target.id in self.symtable:
+                    saved_syms[tnode.target.id] = copy.deepcopy(self._getsym(tnode.target))
 
-                elif tnode.target.__class__ == ast.Tuple:
-                    target = []
-                    for tval in tnode.target.elts:
-                        mylocals[tval.id] = []
-                        if tval.id in self.symtable:
-                            saved_syms[tval.id] = copy.deepcopy(self.symtable[tval.id])
+            elif tnode.target.__class__ == ast.Tuple:
+                target = []
+                for tval in tnode.target.elts:
+                    if tval.id in self.symtable:
+                        saved_syms[tval.id] = copy.deepcopy(self._getsym(tval))
+        return saved_syms
 
-        for tnode in node.generators:
-            if tnode.__class__ == ast.comprehension:
-                ttype = 'name'
-                if tnode.target.__class__ == ast.Name:
-                    if (not valid_symbol_name(tnode.target.id) or
-                        tnode.target.id in self.readonly_symbols):
-                        errmsg = f"invalid symbol name (reserved word?) {tnode.target.id}"
-                        self.raise_exception(tnode.target, exc=NameError, msg=errmsg)
-                    ttype, target = 'name', tnode.target.id
-                elif tnode.target.__class__ == ast.Tuple:
-                    ttype = 'tuple'
-                    target =tuple([tval.id for tval in tnode.target.elts])
 
-                for val in self.run(tnode.iter):
-                    if ttype == 'name':
-                        self.symtable[target] = val
-                    else:
-                        for telem, tval in zip(target, val):
-                            self.symtable[target] = val
+    def do_generator(self, gnodes, node, out):
+        gnode = gnodes[0]
+        nametype = True
+        if gnode.target.__class__ == ast.Name:
+            if (not valid_symbol_name(gnode.target.id) or
+                gnode.target.id in self.readonly_symbols):
+                errmsg = f"invalid symbol name (reserved word?) {gnode.target.id}"
+                self.raise_exception(gnode.target, exc=NameError, msg=errmsg)
+            target = gnode.target.id
+        elif gnode.target.__class__ == ast.Tuple:
+            nametype = False
+            target = tuple([gval.id for gval in gnode.target.elts])
 
-                    add = True
-                    for cond in tnode.ifs:
-                        add = add and self.run(cond)
-                    if add:
-                        if ttype == 'name':
-                            mylocals[target].append(val)
-                        else:
-                            for telem, tval in zip(target, val):
-                                mylocals[telem].append(tval)
-        return mylocals, saved_syms
+        for val in self.run(gnode.iter):
+            if nametype:
+                self.symtable[target] = val
+            else:
+                for telem, tval in zip(target, val):
+                    self.symtable[telem] = tval
+            add = True
+            for cond in gnode.ifs:
+                add = add and self.run(cond)
+                if not add:
+                    break
+            if add:
+                if len(gnodes) > 1:
+                    self.do_generator(gnodes[1:], node, out)
+                elif isinstance(out, list):
+                    out.append(self.run(node.elt))
+                elif isinstance(out, dict):
+                    out[self.run(node.key)] = self.run(node.value)
 
     def on_listcomp(self, node):
-        """List comprehension"""
-        mylocals, saved_syms = self.comprehension_data(node)
-
-        names = list(mylocals.keys())
-        data = list(mylocals.values())
-        def listcomp_recurse(out, i, names, data):
-            if i == len(names):
-                out.append(self.run(node.elt))
-                return
-
-            for val in data[i]:
-                self.symtable[names[i]] = val
-                listcomp_recurse(out, i+1, names, data)
+        """List comprehension v2"""
+        saved_syms = self._comp_save_syms(node)
 
         out = []
-        listcomp_recurse(out, 0, names, data)
+        self.do_generator(node.generators, node, out)
         for name, val in saved_syms.items():
             self.symtable[name] = val
         return out
@@ -826,28 +809,15 @@ class Interpreter:
         return set(self.on_listcomp(node))
 
     def on_dictcomp(self, node):
-        """Dictionary comprehension"""
-        mylocals, saved_syms = self.comprehension_data(node)
-
-        names = list(mylocals.keys())
-        data = list(mylocals.values())
-
-        def dictcomp_recurse(out, i, names, data):
-            if i == len(names):
-                out[self.run(node.key)] = self.run(node.value)
-                return
-
-            for val in data[i]:
-                self.symtable[names[i]] = val
-                dictcomp_recurse(out, i+1, names, data)
+        """Dict comprehension v2"""
+        saved_syms = self._comp_save_syms(node)
 
         out = {}
-        dictcomp_recurse(out, 0, names, data)
+        self.do_generator(node.generators, node, out)
         for name, val in saved_syms.items():
             self.symtable[name] = val
         return out
 
-
     def on_excepthandler(self, node):  # ('type', 'name', 'body')
         """Exception handler..."""
         return (self.run(node.type), node.name, node.body)
@@ -888,17 +858,15 @@ class Interpreter:
         msg = ' '.join(out.args)
         msg2 = self.run(msgnode)
         if msg2 not in (None, 'None'):
-            msg = "%s: %s" % (msg, msg2)
+            msg = f"{msg:s}: {msg2:s}"
         self.raise_exception(None, exc=out.__class__, msg=msg, expr='')
 
     def on_call(self, node):
         """Function execution."""
-        #  ('func', 'args', 'keywords'. Py<3.5 has 'starargs' and 'kwargs' too)
         func = self.run(node.func)
         if not hasattr(func, '__call__') and not isinstance(func, type):
             msg = f"'{func}' is not callable!!"
             self.raise_exception(node, exc=TypeError, msg=msg)
-
         args = [self.run(targ) for targ in node.args]
         starargs = getattr(node, 'starargs', None)
         if starargs is not None:
@@ -963,20 +931,14 @@ class Interpreter:
         args = [tnode.arg for tnode in node.args.args[:offset]]
         doc = None
         nb0 = node.body[0]
-        # deprecation warning: will become
-        # if isinstance(nb0, ast.Expr) and isinstance(nb0.value, ast.Constant):
-        if isinstance(nb0, ast.Expr) and isinstance(nb0.value, ast.Str):
-            # deprecation warning: will become
-            # doc = nb0.value
-            doc = nb0.value.s
-
+        if isinstance(nb0, ast.Expr) and isinstance(nb0.value, ast.Constant):
+            doc = nb0.value
         varkws = node.args.kwarg
         vararg = node.args.vararg
         if isinstance(vararg, ast.arg):
             vararg = vararg.arg
         if isinstance(varkws, ast.arg):
             varkws = varkws.arg
-
         self.symtable[node.name] = Procedure(node.name, self, doc=doc,
                                              lineno=self.lineno,
                                              body=node.body,
